/*
 * Copyright (c) 2009 Iwata <iwata@quasiquote.org>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <unistd.h>
#include <time.h>
#include <getopt.h>
#include <pwd.h>
#include <fcntl.h>
#include <poll.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/param.h>

#include <expat.h>
#include <curl/curl.h>
#include <curl/types.h>
#include <curl/easy.h>

#include "includes.h"

#ifdef HAVE_SYS_QUEUE_H
#include <sys/queue.h>
#else
#include "compat/sys-queue.h"
#endif

#define TI_BASE_URL     "https://twitter.com"
#define TI_API_BASE_URL "https://api.twitter.com"
#define TI_FRIEND_URL   "/statuses/friends_timeline"
#define TI_PUBLIC_URL   "/statuses/public_timeline"
#define TI_USER_URL     "/statuses/user_timeline"
#define TI_FAVORATE_URL "/favorites/create"
#define TI_DIRECT_URL   "/direct_messages/new"
#define TI_UPDATE_URL   "/statuses/update"
#define TI_RETWEET_URL  "/1/statuses/retweet"

#define TI_PASSWD_LEN 128

struct ti_entry {
	char *id;
	char *screen_name;
	char *text;

	TAILQ_ENTRY(ti_entry) ti_entries;
};
typedef TAILQ_HEAD(ti_tailhead, ti_entry) ti_head;

struct ti_xml_tag {
	char *tag;

	TAILQ_ENTRY(ti_xml_tag) ti_xml_tags;
};
typedef TAILQ_HEAD(ti_xmltail, ti_xml_tag) ti_xmlhead;

typedef struct {
	ti_head *head;
	struct ti_entry *ent;
	ti_xmlhead *xml;
} ti_ctx;

typedef struct {
	char *base_url;
	char *api_base_url;
} ti_base_url;

static void *
ti_xmalloc(size_t size)
{
	void *ptr = malloc(size);
	if (ptr)
		return ptr;
	abort();
}

static void *
ti_xrealloc(void *ptr, size_t size)
{
	void *new_ptr;

	if (ptr) {
		if ((new_ptr = realloc(ptr, size)) == NULL) {
			free(ptr);
			abort();
		}
		return new_ptr;
	} else
		return ti_xmalloc(size);
}

static void *
ti_xcalloc(size_t nmemb, size_t size)
{
	void *ptr = calloc(nmemb, size);

	if (ptr)
		return ptr;
	else
		abort();
}

static int
ti_valid_msg(struct ti_entry *ent)
{
	return ent->id && ent->screen_name && ent->text;
}

static int
ti_xml_pathcmp(ti_xmlhead *head, const char *path, ...)
{
	va_list ap;
	const char *p;
	struct ti_xml_tag *np;
	int ret = 0;

	if (!path || (np = TAILQ_LAST(head, ti_xmltail)) == NULL)
		return 0;
	if (strcmp(np->tag, path) != 0)
		return 0;
	va_start(ap, path);
	for ( ; ; ){
		np = TAILQ_PREV(np, ti_xmltail, ti_xml_tags);
		p = va_arg(ap, const char *);
		if (!p && !np) {
			ret = 1;
			break;
		}
		if (!p || !np) {
			ret = 0;
			break;
		}
		if (strcmp(np->tag, p) != 0) {
			ret = 0;
			break;
		}
	}
	va_end(ap);
	return ret;
}

static void
ti_element_start_cb(void *userData, const XML_Char *name, const XML_Char *atts[])
{
	ti_ctx *ctx = userData;
	struct ti_xml_tag *tag = ti_xcalloc(1, sizeof(struct ti_xml_tag));

	tag->tag = strdup(name);
	TAILQ_INSERT_HEAD(ctx->xml, tag, ti_xml_tags);

	if (ti_xml_pathcmp(ctx->xml, "statuses", "status", NULL))
		ctx->ent = ti_xcalloc(1, sizeof(*ctx->ent));
}

static void
ti_element_end_cb(void *userData, const XML_Char *name)
{
	ti_ctx *ctx = userData;
	struct ti_xml_tag *np;

	if ((np = TAILQ_FIRST(ctx->xml)) != NULL) {
		if (strcmp(np->tag, name) != 0)
			fprintf(stderr, "xml data corruption\n");

		if (ti_xml_pathcmp(ctx->xml, "statuses", "status", NULL)) {
			if (ti_valid_msg(ctx->ent))
				TAILQ_INSERT_HEAD(ctx->head, ctx->ent, ti_entries);
		}

		TAILQ_REMOVE(ctx->xml, np, ti_xml_tags);
		free(np->tag);
		free(np);
	} else
		fprintf(stderr, "unopend xml tag\n");
}

static int
ti_strappend(char **dest, const char *append, size_t append_len)
{
	if (*dest) {
		char *str;
		size_t dest_len = strlen(*dest);
		size_t len = dest_len + append_len;

		str = ti_xrealloc(*dest, len + 1);
		memcpy(&str[dest_len], append, append_len);
		str[len] = '\0';
		*dest = str;
	} else {
		char *str;

		str = ti_xmalloc(append_len + 1);
		memcpy(str, append, append_len);
		str[append_len] = '\0';
		*dest = str;
	}
	return 1;
}

static void
ti_characterdata_cb(void *userData, const XML_Char *s, int len)
{
	ti_ctx *ctx = userData;

	if (ti_xml_pathcmp(ctx->xml, "statuses", "status", "id", NULL))
		ti_strappend(&ctx->ent->id, s, len);

	if (ti_xml_pathcmp(ctx->xml, "statuses", "status", "text", NULL))
		ti_strappend(&ctx->ent->text, s, len);

	if (ti_xml_pathcmp(ctx->xml, "statuses", "status", "user", "screen_name", NULL))
		ti_strappend(&ctx->ent->screen_name, s, len);

}

static int
ti_write_entry(ti_head *head, const char *prefix)
{
	struct ti_entry *np;

	TAILQ_FOREACH(np, head, ti_entries) {
		char path[MAXPATHLEN];
		char msg[1024];
		int out;

		snprintf(path, sizeof(path), "%s/out", prefix);
		if ((out = open(path, O_WRONLY | O_APPEND, S_IWUSR)) == -1) {
			perror("open");
			exit(EXIT_FAILURE);
		}
		snprintf(msg, sizeof(msg), "%s: %s #%s\n", np->screen_name, np->text, np->id);
		write(out, msg, strlen(msg));
		close(out);
	}

	return 1;
}

static void
ti_free_entry(ti_head *head)
{
	struct ti_entry *np;

	while ((np = TAILQ_FIRST(head)) != NULL) {
		TAILQ_REMOVE(head, np, ti_entries);
		free(np->id);
		free(np->screen_name);
		free(np->text);
		free(np);
	}
}


static size_t
ti_get_since_id(char *since_id, ti_head *head, size_t len)
{
	struct ti_entry *np;

	np = TAILQ_LAST(head, ti_tailhead);
	return strlcpy(since_id, np->id, len);
}

static int
ti_process_xml(ti_head *head, char *xml_str)
{
	XML_Parser parser;
	ti_ctx ctx;
	ti_xmlhead xml;
	struct ti_xml_tag *np;

	if ((parser = XML_ParserCreate("UTF-8")) == NULL) {
		fprintf(stderr, "parser creation error\n");
		exit(1);
	}

	TAILQ_INIT(&xml);

	ctx.ent = NULL;
	ctx.xml = &xml;
	ctx.head = head;
	XML_SetUserData(parser, &ctx);
	XML_SetElementHandler(parser, ti_element_start_cb, ti_element_end_cb);
	XML_SetCharacterDataHandler(parser, ti_characterdata_cb);
	XML_Parse(parser, xml_str, strlen(xml_str), 1);
	XML_ParserFree(parser);

	while ((np = TAILQ_FIRST(&xml)) != NULL) {
		TAILQ_REMOVE(&xml, np, ti_xml_tags);
		free(np->tag);
		free(np);
	}

	return 0;
}

/*
 * code from example/getinmemory.c in libcurl
 */
struct ti_mem {
	char *memory;
	size_t size;
};

static size_t
ti_fetch_cb(void *ptr, size_t size, size_t nmemb, void *data)
{
	size_t realsize = size * nmemb;
	struct ti_mem *mem = (struct ti_mem *)data;

	mem->memory = ti_xrealloc(mem->memory, mem->size + realsize + 1);
	if (mem->memory) {
		memcpy(&(mem->memory[mem->size]), ptr, realsize);
		mem->size += realsize;
		mem->memory[mem->size] = 0;
	}
	return realsize;
}

static int
ti_fetch_with_default(char **ret_str, CURL *curl_handle, const char *url)
{
	struct ti_mem chunk = { NULL, 0 };
	CURLcode curl_ret;
	int ret = 1;
	char *proxy;

	curl_easy_setopt(curl_handle, CURLOPT_URL, url);
	curl_easy_setopt(curl_handle, CURLOPT_WRITEFUNCTION, ti_fetch_cb);
	curl_easy_setopt(curl_handle, CURLOPT_WRITEDATA, (void *)&chunk);

	curl_easy_setopt(curl_handle, CURLOPT_SSL_VERIFYPEER, 0);
	curl_easy_setopt(curl_handle, CURLOPT_SSL_VERIFYHOST, 1);

	if ((proxy = getenv("http_proxy")) != NULL)
		curl_easy_setopt(curl_handle, CURLOPT_PROXY, proxy);

	curl_ret = curl_easy_perform(curl_handle);

	if (curl_ret != CURLE_OK) {
		printf("libcurl: %s\n", curl_easy_strerror(curl_ret));
		ret = 0;
	}

	*ret_str = chunk.memory;

	return ret;
}

static int
ti_fetch(char **ret_str, const char *url, const char *user, const char *passwd)
{
	CURL *curl_handle;
	int ret = 1;

	curl_global_init(CURL_GLOBAL_ALL);
	curl_handle = curl_easy_init();

	if (user) {
		char auth_str[1024];

		snprintf(auth_str, sizeof(auth_str), "%s:%s", user, passwd);
		curl_easy_setopt(curl_handle, CURLOPT_HTTPAUTH, (long)CURLAUTH_BASIC);
		curl_easy_setopt(curl_handle, CURLOPT_USERPWD, auth_str);
	}

	ret = ti_fetch_with_default(ret_str, curl_handle, url);

	curl_easy_cleanup(curl_handle);

	curl_global_cleanup();

	return ret;
}

static int
ti_send_update(const char *buf, const ti_base_url *base_url, const char *user, const char *passwd, const char *extra)
{
	CURL *curl_handle;
	char auth_str[1024];
	int ret = 1;
	char url[1024], *escaped_msg, msg[1024];
	char *ret_str;

	if (!user)
		return 0;

	curl_global_init(CURL_GLOBAL_ALL);
	curl_handle = curl_easy_init();

	snprintf(url, sizeof(url), "%s%s.xml", base_url->base_url, TI_UPDATE_URL);

	escaped_msg = curl_easy_escape(curl_handle, buf, 0);
	snprintf(msg, sizeof(msg), "status=%s&source=API%s", escaped_msg, extra);
	curl_free(escaped_msg);
	curl_easy_setopt(curl_handle, CURLOPT_POSTFIELDS, msg);

	snprintf(auth_str, sizeof(auth_str), "%s:%s", user, passwd);
	curl_easy_setopt(curl_handle, CURLOPT_HTTPAUTH, (long)CURLAUTH_BASIC);
	curl_easy_setopt(curl_handle, CURLOPT_USERPWD, auth_str);

	ret = ti_fetch_with_default(&ret_str, curl_handle, url);

	curl_easy_cleanup(curl_handle);

	curl_global_cleanup();

	/* XXX */
	free(ret_str);

	return ret;
}

static int
ti_send_favorate(const char *user, const ti_base_url *base_url, const char *passwd, const char *id)
{
	CURL *curl_handle;
	char auth_str[1024];
	int ret = 1;
	char *ret_str;
	char url[1024], msg[1024];

	if (!user)
		return 0;

	curl_global_init(CURL_GLOBAL_ALL);
	curl_handle = curl_easy_init();

	snprintf(url, sizeof(url), "%s%s/%s.xml", base_url->base_url, TI_FAVORATE_URL, id);

	snprintf(msg, sizeof(msg), "id=%s&source=API", id);
	curl_easy_setopt(curl_handle, CURLOPT_POSTFIELDS, msg);

	snprintf(auth_str, sizeof(auth_str), "%s:%s", user, passwd);
	curl_easy_setopt(curl_handle, CURLOPT_HTTPAUTH, (long)CURLAUTH_BASIC);
	curl_easy_setopt(curl_handle, CURLOPT_USERPWD, auth_str);

	ret = ti_fetch_with_default(&ret_str, curl_handle, url);

	curl_easy_cleanup(curl_handle);

	curl_global_cleanup();

	/* XXX */
	free(ret_str);

	return ret;
}

static int
ti_send_direct(const ti_base_url *base_url, const char *from_user, const char *passwd, const char *direct_msg, const char *to_user)
{
	CURL *curl_handle;
	char auth_str[1024];
	int ret = 1;
	char *ret_str;
	char url[1024], *escaped_msg, msg[1024];

	if (!from_user || !to_user)
		return 0;

	curl_global_init(CURL_GLOBAL_ALL);
	curl_handle = curl_easy_init();

	snprintf(url, sizeof(url), "%s%s.xml", base_url->base_url, TI_DIRECT_URL);

	escaped_msg = curl_easy_escape(curl_handle, direct_msg, 0);
	snprintf(msg, sizeof(msg), "text=%s&user=%s", escaped_msg, to_user);
	curl_free(escaped_msg);
	curl_easy_setopt(curl_handle, CURLOPT_POSTFIELDS, msg);

	snprintf(auth_str, sizeof(auth_str), "%s:%s", from_user, passwd);
	curl_easy_setopt(curl_handle, CURLOPT_HTTPAUTH, (long)CURLAUTH_BASIC);
	curl_easy_setopt(curl_handle, CURLOPT_USERPWD, auth_str);

	ret = ti_fetch_with_default(&ret_str, curl_handle, url);

	curl_easy_cleanup(curl_handle);

	curl_global_cleanup();

	/* XXX */
	free(ret_str);

	return ret;
}

static int
ti_send_retweet(const char *user, const ti_base_url *base_url, const char *passwd, const char *id)
{
	CURL *curl_handle;
	char auth_str[1024];
	int ret = 1;
	char *ret_str;
	char url[1024], msg[1024];

	if (!user)
		return 0;

	curl_global_init(CURL_GLOBAL_ALL);
	curl_handle = curl_easy_init();

	snprintf(url, sizeof(url), "%s%s/%s.xml", base_url->api_base_url, TI_RETWEET_URL, id);

	snprintf(msg, sizeof(msg), "id=%s&source=API", id);
	curl_easy_setopt(curl_handle, CURLOPT_POSTFIELDS, msg);

	snprintf(auth_str, sizeof(auth_str), "%s:%s", user, passwd);
	curl_easy_setopt(curl_handle, CURLOPT_HTTPAUTH, (long)CURLAUTH_BASIC);
	curl_easy_setopt(curl_handle, CURLOPT_USERPWD, auth_str);

	ret = ti_fetch_with_default(&ret_str, curl_handle, url);

	curl_easy_cleanup(curl_handle);

	curl_global_cleanup();

	/* XXX */
	free(ret_str);

	return ret;
}

static int
ti_make_io_dir(char *prefix)
{
	char path[MAXPATHLEN];

	mkdir(prefix, S_IRWXU);

	snprintf(path, sizeof(path), "%s/in", prefix);
	unlink(path);
	mkfifo(path, S_IRWXU);

	snprintf(path, sizeof(path), "%s/out", prefix);
	unlink(path);
	mkfifo(path, S_IRWXU);

	return 1;
}

static int
ti_open_input_file(const char *prefix)
{
	char path[MAXPATHLEN];
	int ifd;

	snprintf(path, sizeof(path), "%s/in", prefix);
	if ((ifd = open(path, O_RDONLY | O_NONBLOCK, S_IWUSR)) == -1) {
		perror("ti_open_io_files(\"in\")");
		exit(EXIT_FAILURE);
	}
	return ifd;
}

static int
ti_socket_ready(int fd, int timeout)
{
	struct pollfd pfd;
	int ndfs;

	pfd.fd = fd;
	pfd.events = POLLIN;
	ndfs = poll(&pfd, 1, timeout);

	if (ndfs < 0) {
		perror("poll");
		return 0;
	} else if (ndfs == 0)
		return 0;
	else
		return 1;
}

static int
ti_read_line(int fd, char *buf, size_t len)
{
	unsigned char c;
	size_t i = 0;

	do {
		if (read(fd, &c, 1) == 0 || c == '\n') {
			buf[i] = '\0';
			break;
		}
		buf[i++] = c;
	} while (i < len);
	return i;
}

static int
ti_get_last_id(char *id, size_t id_len, const ti_base_url *base_url, const char *friend, const char *user, const char *passwd)
{
	char *buf;
	char url[1024];
	ti_head head;

	snprintf(url, sizeof(url), "%s%s/%s.xml?count=1", base_url->base_url, TI_USER_URL, friend);
	if (!ti_fetch(&buf, url, user, passwd))
		printf("error");
	TAILQ_INIT(&head);
	ti_process_xml(&head, buf);
	free(buf);
	if (TAILQ_EMPTY(&head))
		return 0;
	else {
		struct ti_entry *np = TAILQ_FIRST(&head);
		strlcpy(id, np->id, id_len);
	}
	ti_free_entry(&head);
	return 1;
}

static int
ti_is_responce_with_at(char *dest, size_t len, const char *s, char **friend)
{
	char *p;

	if (s[0] != '@')
		return 0;
	strlcpy(dest, &s[1], len);
	*friend = dest;
	p = dest;
	if ((p = strchr(dest, ' ')) != NULL)
		*p = '\0';
	return 1;
}

static int
ti_is_responce_with_id(char *dest, size_t len, const char *s, char **id, char **msg)
{
	if (s[0] != '#')
		return 0;
	strlcpy(dest, &s[1], len);
	if ((*id = strtok(dest, " \t")) == NULL)
		return 0;
	if ((*msg = strtok(NULL, "")) == NULL)
		return 0;
	return 1;
}

static int
ti_is_favorate(char *dest, size_t len, const char *s, char **friend)
{
	char *is_fav;

	if (s[0] != 'f')
		return 0;
	strlcpy(dest, s, len);
	is_fav = strtok(dest, " \t");
	if (strcasecmp(is_fav, "fav") != 0)
		return 0;
	if ((*friend = strtok(NULL, " \t")) == NULL)
		return 0;
	return 1;
}

static int
ti_is_direct(char *dest, size_t len, const char *s, char **friend, char **msg)
{
	char *is_d;

	if (s[0] != 'D' && s[0] != 'd')
		return 0;
	strlcpy(dest, s, len);
	is_d = strtok(dest, " \t");
	if (strcasecmp(is_d, "D") != 0)
		return 0;
	if ((*friend = strtok(NULL, " \t")) == NULL)
		return 0;
	if ((*msg = strtok(NULL, "")) == NULL)
		return 0;
	return 1;
}

static int
ti_is_retweet(char *dest, size_t len, const char *s, char **id)
{
	char *is_rt;

	if (s[0] != 'R')
		return 0;
	strlcpy(dest, s, len);
	is_rt = strtok(dest, " \t");
	if (strcasecmp(is_rt, "RT") != 0)
		return 0;
	if ((*id = strtok(NULL, " \t")) == NULL)
		return 0;
	if (*id[0] != '#')
		return 0;
	*id = *id + 1;
	return 1;
}

static int
ti_process_cmd(int ifd, const ti_base_url *base_url, const char *user, const char *passwd)
{
	char line[1024];
	char buf[256];
	char id[1024];
	char resp[256];
	char *friend;
	char *send_to, *msg;
	char *send_id;

	ti_read_line(ifd, line, sizeof(line));

	if (ti_is_responce_with_at(buf, sizeof(buf), line, &friend)) {
		if (ti_get_last_id(id, sizeof(id), base_url, friend, user, passwd)) {
			snprintf(resp, sizeof(resp), "&in_reply_to_status_id=%s", id);
			ti_send_update(line, base_url, user, passwd, resp);
		} else
			ti_send_update(line, base_url, user, passwd, "");
	} else if (ti_is_responce_with_id(buf, sizeof(buf), line, &send_id, &msg)) {
		snprintf(resp, sizeof(resp), "&in_reply_to_status_id=%s", send_id);
		ti_send_update(msg, base_url, user, passwd, resp);
	} else if (ti_is_favorate(buf, sizeof(buf), line, &friend)) {
		if (ti_get_last_id(id, sizeof(id), base_url, friend, user, passwd))
			ti_send_favorate(user, base_url, passwd, id);
	} else if (ti_is_direct(buf, sizeof(buf), line, &send_to, &msg))
		ti_send_direct(base_url, user, passwd, msg, send_to);
	else if (ti_is_retweet(buf, sizeof(buf), line, &send_id))
		ti_send_retweet(user, base_url, passwd, send_id);
	else
		ti_send_update(line, base_url, user, passwd, "");

	return 1;
}

static int
ti_read_keyfile(const char *keyfile, char **user, char **passwd)
{
	FILE *fp;
	char buf[1024];
	char *p;

	if ((fp = fopen(keyfile, "r")) == NULL) {
		perror("ti_read_passfile");
		return 0;
	}

	if (fgets(buf, sizeof(buf), fp) != NULL)
		buf[strcspn(buf, "\n")] = '\0';

	if ((p = strchr(buf, ':')) == NULL) {
		fprintf(stderr, "password format error (should 'user:passwd').\n");
		fclose(fp);
		return 0;
	}

	buf[strcspn(buf, ":")] = '\0';

	*user = strdup(buf);
	*passwd = strdup(p + 1);

	fclose(fp);
	return 1;
}

static void
usage(void)
{
	printf("ti [-b baseurl] [-k passwd] [-f passwdfile] [-t prefix] [-u user] [-w interval]\n");
	exit(EXIT_FAILURE);
}

int
main(int argc, char *argv[])
{
	char *buf;
	ti_head head;
	int ch;
	int ifd;
	char *user = NULL, *passwd = NULL;
	char *keyfile = NULL;
	char *prefix = NULL;
	int interval = 60;
	char since_id[1024];
	ti_base_url base_url = { TI_BASE_URL, TI_API_BASE_URL };
	const char *errstr;
	int i;

	while ((ch = getopt(argc, argv, "b:B:k:f:t:u:w:")) != -1) {
		switch (ch) {
		case 'b':
			base_url.base_url = optarg;
		case 'B':
			base_url.api_base_url = optarg;
		case 'k':
			passwd = optarg;
			break;
		case 'f':
			keyfile = optarg;
			break;
		case 't':
			prefix = optarg;
			break;
		case 'u':
			user = optarg;
			break;
		case 'w':
			interval = (int)strtonum(optarg, 1, 100000, &errstr);
                        if (errstr) {
				printf("number of iterations is %s: %s\n", errstr, optarg);
                                usage();
                        }
                        break;
		default:
			usage();
		}
	}

	if (keyfile)
		if (!ti_read_keyfile(keyfile, &user, &passwd))
			return 1;

	if (user && !passwd)
		passwd = getpass("Password: ");

	if (!prefix) {
		struct passwd *pw;

		pw = getpwuid(getuid());
		asprintf(&prefix, "%s/ti", pw->pw_dir);
		endpwent();
	}

	ti_make_io_dir(prefix);
	ifd = ti_open_input_file(prefix);

	TAILQ_INIT(&head);
	since_id[0] = '\0';

	while (1) {
		char url[1024];

		if (!user) {
			if (since_id[0] == '\0')
				snprintf(url, sizeof(url), "%s%s.xml?count=200", base_url.base_url, TI_PUBLIC_URL);
			else
				snprintf(url, sizeof(url), "%s%s.xml?since_id=%s&count=200", base_url.base_url, TI_PUBLIC_URL, since_id);
		} else {
			if (since_id[0] == '\0')
				snprintf(url, sizeof(url), "%s%s/%s.xml?count=200", base_url.base_url, TI_FRIEND_URL, user);
			else
				snprintf(url, sizeof(url), "%s%s/%s.xml?since_id=%s&count=200", base_url.base_url, TI_FRIEND_URL, user, since_id);
		}
		if (!ti_fetch(&buf, url, user, passwd))
			goto next_fetch;
		ti_process_xml(&head, buf);
		if (!TAILQ_EMPTY(&head)) {
			ti_write_entry(&head, prefix);
			ti_get_since_id(since_id, &head, sizeof(since_id));
			ti_free_entry(&head);
		}
		free(buf);
	next_fetch:

		/* waiting for input while 'interval' sec. */
		for (i = 0; i < interval; i++)
			if (ti_socket_ready(ifd, 1000)) {
				ti_process_cmd(ifd, &base_url, user, passwd);
				break;
			}
	}

	if (!TAILQ_EMPTY(&head))
		ti_free_entry(&head);

	return 1;
}
